package dictionarybuilder;

import java.io.*;

/**
 * 
 *
 * @author quasiyoke
 */
public abstract class EasyAccessFile implements Closeable{

protected boolean writingResolved;

protected long filePointer;

public EasyAccessFile(File file,boolean resolveWriting) throws IOException{
    writingResolved=resolveWriting;
}

public EasyAccessFile(String filePath,boolean resolveWriting) throws
	IOException{
    writingResolved=resolveWriting;
}

protected void finalize(){
    try{
	close();
    }catch(IOException e){
    }
}

public long getFilePointer(){
    return filePointer;
}

public abstract char readChar() throws IOException;

public abstract int readInt() throws IOException;

public abstract int readUnsignedByte() throws IOException;

/**
 * @return Прочитанный символ.
 * @exception EOFException Если поток достигает конца до прочтения всех
 * нужных байт.
 * @exception IOException Когда происходит ошибка ввода/вывода.
 */
public char readUTF8Char() throws IOException{
    int character;
    int nByte=readUnsignedByte();
    if((nByte&0x80)==0)
	/*
	 * Символы Unicode из диапазона '\u0000' —'\u007F', в котором лежит
	 * английский алфавит, записываются одним байтом, старший байт
	 * просто отбрасывается.
	 */
	return (char)nByte;
    else{
	if((nByte&0xE0)==0xC0)
	    /*
	     * Символы Unicode из диапазона '\u0080' —'\u07FF', в котором лежат
	     * наиболее распространенные символы национальных алфавитов,
	     * записываются двумя байтами следующим образом: символ Unicode с
	     * кодировкой 00000хххххуууууу записывается как 110ххххх10уууууу.
	     */
	    character=(nByte&0x1F)<<6;
	else
	    /*
	     * Остальные символы Unicode из диапазона '\u0800' —'\UFFFF'
	     * записываются тремя байтами по следующему правилу: символ Unicode
	     * с кодировкой xxxxyyyyyyzzzzzz записывается как
	     * 1110xxxx10yyyyyy10zzzzzz.
	     */
	    character=((nByte&0x0F)<<12)|(readUnsignedByte()&0x3F)<<6;
	return (char)(character|(readUnsignedByte()&0x3F));
    }
}

public String readUTF8String() throws IOException{
    StringBuffer stringBuffer=new StringBuffer();
    char character;
    while((character=readUTF8Char())<=0x000D);
    try{
	do
	    stringBuffer.append(character);
	while((character=readUTF8Char())>0x000D);
    }catch(EOFException e){
    }
    return stringBuffer.toString();
}

/**
 * Выполняет перемещение указателя в указанное положение. Никогда не вызывает
 * {@link EOFException}: в случае выхода за пределы файла, размер его изменится
 * во время первой же записывающей операции, а в пропущенное новое пространство
 * попадут неопределённые данные.
 */
public abstract void seek(long position) throws IOException;

/**
 * Выполняет перемещение указателя вперёд на учазанное число байт. Если
 * <code>bytesCount</code> меньше единицы, переход не осуществляется.
 */
public abstract void skipBytes(int bytesCount) throws IOException;

/**
 * Пытается прочесть первые байты файла, и если они не составляют BOM-сигнатуру
 * 0xEF, 0xBB, 0xBF; или возникает {@link EOFException} (файл меньше трёх байт),
 * переходит в исходную позицию.
 */
public void skipBOM() throws IOException{
    long position=getFilePointer();
    try{
	if(readUTF8Char()!=0xFEFF)
	    seek(position);
    }catch(EOFException e){
	seek(position);
    }
}

public abstract void writeByte(int nByte) throws IOException;

public abstract void writeChar(char character) throws IOException;

public abstract void writeInt(int number) throws IOException;

/**
 * @return Количество байт, которое было записано.
 */
public int writeUTF8Char(char character) throws IOException{
    if(character<=0x007F){
	/*
	 * Символы Unicode из диапазона '\u0000' —'\u007F', в котором лежит
	 * английский алфавит, записываются одним байтом, старший байт
	 * просто отбрасывается.
	 */
	writeByte(character);
	return 1;
    }
    int nCharacter;
    int bytesCount;
    if(character<=0x07FF){
	/*
	 * Символы Unicode из диапазона '\u0080' —'\u07FF', в котором лежат
	 * наиболее распространенные символы национальных алфавитов,
	 * записываются двумя байтами следующим образом: символ Unicode с
	 * кодировкой 00000хххххуууууу записывается как 110ххххх10уууууу.
	 */
	bytesCount=2;
	nCharacter=0xC0|(character>>6);
	writeByte(nCharacter);
    }else{
	/*
	 * Остальные символы Unicode из диапазона '\u0800' —'\UFFFF'
	 * записываются тремя байтами по следующему правилу: символ Unicode
	 * с кодировкой xxxxyyyyyyzzzzzz записывается как
	 * 1110xxxx10yyyyyy10zzzzzz.
	 */
	bytesCount=3;
	nCharacter=0xE0|(character>>12);
	writeByte(nCharacter);
	nCharacter=0x80|((character>>6)&0x3F);
	writeByte(nCharacter);
    }
    nCharacter=0x80|(character&0x3F);
    writeByte(nCharacter);
    return bytesCount;
}

/**
 * @return Число записанных в поток байт.
 */
public int writeUTF8String(String string)
	throws IOException{
    int j;
    char character;
    final char END_OF_STRING_CHARACTER=0;
    int stringLength=string.length();
    int bytesCount=0;
    for(j=0;j<stringLength;j++)
	if((character=string.charAt(j))!=END_OF_STRING_CHARACTER)///////////////////////////////////////////////
	    bytesCount+=writeUTF8Char(character);
    bytesCount+=writeUTF8Char(END_OF_STRING_CHARACTER);
    return bytesCount;
}

}
