import java.io.*;
import java.util.*;
import javax.microedition.lcdui.*;

/**
 * 
 *
 * @author quasiyoke
 */
public class Dictionary implements Comparable{

private final static int NULL_INT_VALUE=-2;

private char[] alphabet;

private EasyReadingFilesColony dataInput;

private String keysLanguage;

private String articlesLanguage;

private String source;

private char[][] keysChars;

private Index root;

private char[] sortedAlphabet;

/**
 * @param keysChars Соответствие ряда символов кнопкам, на которых они
 * расположены. {@code keysChars[2]} -- символы, закреплённые за клавишей 2.
 */
public Dictionary(String path) throws IOException{
    dataInput=new EasyReadingFilesColony(path);
    keysLanguage=dataInput.readUTF8String('\0');
    articlesLanguage=dataInput.readUTF8String('\0');
    source=dataInput.readUTF8String('\0');
    int alphabetLength=dataInput.readUnsignedByte();
    alphabet=new char[alphabetLength];
    sortedAlphabet=new char[alphabetLength];
    int j;
    for(j=0;j<alphabetLength;j++){
        alphabet[j]=sortedAlphabet[j]=
                Character.toLowerCase(dataInput.readChar());
    }
    sortedAlphabet=PieceOfDictionary.sort(sortedAlphabet);
    int rootSubindexesCount=dataInput.readUnsignedByte();
    root=
            new Index((int)dataInput.getFilePointer(),'\0',rootSubindexesCount,
            null);
}

public int compareTo(Object object){
    if(object instanceof Dictionary){
        return getKeysLanguage().compareTo(
                ((Dictionary)object).getKeysLanguage());
    }
    return 1;
}

private String getArticlesLanguage(){
    return articlesLanguage;
}

public String getKeysLanguage(){
    return keysLanguage;
}

public char[][] getKeysChars(){
    return keysChars;
}

private String getSource(){
    return source;
}

public char[][] matchKeysCharsAndAlphabet(char[][] keysChars){
    this.keysChars=new char[keysChars.length][];
    int i;
    int j;
    char character;
    StringBuffer keyChars;
    for(j=0;j<keysChars.length;j++){
        keyChars=new StringBuffer();
        for(i=0;i<keysChars[j].length;i++){
            if(PieceOfDictionary.search(
                    sortedAlphabet,character=keysChars[j][i])!=-1){
                keyChars.append(character);
            }
        }
        this.keysChars[j]=keyChars.toString().toCharArray();
    }
    return this.keysChars;
}

public String toString(){
    return Language.getCurrentLanguageInscription("language_"+getKeysLanguage())+
            " -- "+Language.getCurrentLanguageInscription("language_"+
            getArticlesLanguage())+" ("+getSource()+")";
}

class Head implements Comparable{

private Index index;

private String key;

private int position=0;

private int size=NULL_INT_VALUE;

private Head(Index head,String key){
    this.index=head;
    this.key=key;
}

public int compareTo(Object object){
    if(object instanceof Head){
        return key.compareTo(((Head)object).key);
    }
    return 1;
}

public Article elementAt(int[] arIndex,int articleKeyLength) throws
        IndexOutOfSomeBoundsException{
    return index.getArticle(arIndex,articleKeyLength,this);
}

public int size(){
    return size==NULL_INT_VALUE?size=index.getArticlesCount():size;
}

class Article{

private String key;

private String value;

private Article(String key,String value){
    this.key=key;
    this.value=value;
}

public String getHeadKey(){
    return Head.this.key;
}

public int getIndex(String string,int lettersCount){
    final int STRING_LENGTH=string.length();
    int retval;
    for(retval=0;retval<STRING_LENGTH&&lettersCount>0;retval++){
        if(!PieceOfDictionary.isSpace(string.charAt(retval))){
            lettersCount--;
        }
    }
    return retval;
}

public String getKey(){
    return key;
}

public String getText(){
    return value;
}

public void print(Graphics g,final Point POSITION,Point screenSize){
    POSITION.x=PieceOfDictionary.MARGIN_LEFT;
    String key=getKey();
    int keyChainIndex=getIndex(key,MainCanvas.getInstance().getKeyChainLength());
    PieceOfDictionary.print(g,key.substring(0,keyChainIndex),POSITION,
            PieceOfDictionary.FONT_ARTICLES_KEYS_HIGHLIGHTED,
            PieceOfDictionary.COLOR_ARTICLES_KEYS,screenSize);
    PieceOfDictionary.print(g,key.substring(keyChainIndex),POSITION,
            PieceOfDictionary.FONT_ARTICLES_KEYS,
            PieceOfDictionary.COLOR_ARTICLES_KEYS,screenSize);
    PieceOfDictionary.print(g,' '+getText(),POSITION,
            PieceOfDictionary.FONT_ARTICLES_TEXT,
            PieceOfDictionary.COLOR_ARTICLES_TEXT,screenSize);
    POSITION.y+=PieceOfDictionary.FONT_ARTICLES_TEXT.getHeight();
    g.setColor(PieceOfDictionary.COLOR_BACKGROUND);
    g.fillRect(0,POSITION.y,screenSize.x,
            PieceOfDictionary.ARTICLES_DIVIDER_HEIGHT);
    POSITION.y+=PieceOfDictionary.ARTICLES_DIVIDER_HEIGHT;
}

}

}

class HeadsSet{

private Vector heads=new Vector();

private Vector freeHeads=new Vector();

public HeadsSet(Vector keyChain,int keyChainLength){
    heads=new Vector();
    freeHeads=new Vector();
    root.findHeads(keyChain,keyChainLength,heads,freeHeads);
}

/**
 * @throws NoSuchElementException Если статья с заданным индексом не найдена,
 * но ещё есть более "глубокие" статьи.
 * @throws IndexOutOfSomeBoundsException Если нет статьи с заданным индексом.
 */
public Head.Article elementAt(int[] arIndex,int keyLength) throws
        IndexOutOfSomeBoundsException{
    if(arIndex[0]<0){
        throw new IndexOutOfSomeBoundsException();
    }
    int j;
    Vector heads=this.heads;
    int failuresCount=0;
    for(j=0;;j++){
        if(j==heads.size()){
            if(heads==this.heads){
                heads=this.freeHeads;
                j=-1;

            }else{
                if(failuresCount==this.heads.size()+this.freeHeads.size()){
                    throw new IndexOutOfSomeBoundsException();
                }else{
                    throw new NoSuchElementException();
                }
            }
        }else{
            try{
                return ((Head)heads.elementAt(j)).elementAt(arIndex,keyLength);
            }catch(NoSuchElementException e){
            }catch(IndexOutOfSomeBoundsException e){
                failuresCount++;
            }
        }
    }
}

}

class Index implements Comparable{

/**
 * Символ, использующийся для обозначения индекса, которому принадлежит
 * любое количество любых подиндексов.
 */
private final static char FREE_CHARACTER='+';

private final static int HAVE_NOT_FREE_SUBINDEX_MASK=0x80000000;

private final static int NULL_INT_VALUE=-2;

private final static int NULL_POINTER=0xFFFFFFFF;

private final static int SUBINDEXES_OFFSET_MASK=0x7FFFFFFF;

private int articleOffset=NULL_INT_VALUE;

private int articlesCount=NULL_INT_VALUE;

/**
 * Буква, представляемая индексом.
 */
private char character;

private Index freeSubindex;

private String key;

/**
 * Длина ключа к статье, которую представляет этот индекс.
 */
private int articleKeyLength;

/**
 * Как много статей есть у подиндексов без погружения вглубь, в дальнейшую
 * рекурсию. Инициализируется вместе с подиндексами в {@link
 * Index#getSubindexes()}.
 */
private int nextLevelArticlesCount;

private int offset;

private Index parent;

/**
 * Отсортированные подиндексы этого индекса. Заполнется "лениво" в {@link
 * Index#getSubindexes()}.
 */
private Index[] subindexes;

private int subindexesCount;

private int subindexesOffset=NULL_INT_VALUE;

private Index(int offset,char character,int subindexesCount,Index parent){
    this.offset=offset;
    this.character=character;
    this.subindexesCount=subindexesCount;
    this.parent=parent;
    articleKeyLength=parent==null?0:parent.articleKeyLength+1;
}

/**
 * Добавлет в конец данного строкового буфера ключ индекса.
 */
private StringBuffer appendKey(StringBuffer stringBuffer){
    if(parent==null){
        return stringBuffer;
    }
    parent.appendKey(stringBuffer);
    stringBuffer.append(character);
    return stringBuffer;
}

public int compareTo(Object object){
    if(object instanceof Index){
        Index index=(Index)object;
        return index.character==FREE_CHARACTER||index.character>character?-1:index.character<
                character?1:0;
    }
    return 1;
}

private void findHeads(Vector keyChain,int keyChainLength,Vector heads,
        Vector freeHeads){
    if(articleKeyLength==keyChainLength){
        heads.addElement(getHead());
    }else{
        if(freeSubindex!=null){
            freeHeads.addElement(freeSubindex.getHead());
        }
        Index subindex;
        int key=((Character)keyChain.elementAt(articleKeyLength)).charValue();
        if('0'<=key&&key<='9'){
            int j;
            char[] keyChars=keysChars[key-'0'];
            for(j=0;j<keyChars.length;j++){
                if((subindex=findSubindex(keyChars[j]))!=
                        null){
                    subindex.findHeads(keyChain,keyChainLength,heads,
                            freeHeads);
                }
            }
        }else if((subindex=findSubindex((char)key))!=null){
            subindex.findHeads(keyChain,keyChainLength,heads,freeHeads);
        }
    }
}

public Index findSubindex(char character){
    Index[] subindexes=getSubindexes();
    if(subindexes.length==0){
        return null;
    }
    int middle;
    int begin=0;
    int end=subindexes.length-1;
    do{
        middle=begin+end>>1;
        if(character==subindexes[middle].character){
            return subindexes[middle];
        }
        if(character<subindexes[middle].character){
            end=middle-1;
        }else{
            begin=middle+1;
        }
    }while(begin<=end);
    return null;
}

public Head.Article getArticle(Head head){
    if(haveAnArticle()){
        try{
            return head.new Article(dataInput.seek(getArticleTextOffset()).
                    readUTF8String('\1'),dataInput.readUTF8String('\0'));
        }catch(IOException e){
            PieceOfDictionary.showErrorAndExit(Language.
                    getCurrentLanguageInscription("article_text_reading_error"));
        }
    }
    return null;
}

/**
 * @throw NoSuchElementException Когда глубина этого индекса равна требуемой
 * длине ключа, а индекс статьи не имеет.
 * @throw IndexOutOfSomeBoundsException Когда задан индекс статьи,
 * выходящий за рамки имеющихся у индекса статей.
 */
public Head.Article getArticle(final int[] arIndex,final int articleKeyLength,
        Head head) throws IndexOutOfSomeBoundsException{
    if(articleKeyLength==this.articleKeyLength){
        if(haveAnArticle()){
            if(arIndex[0]==0){
                return getArticle(head);
            }else{
                arIndex[0]--;
            }
        }
        if(getArticlesCount()==1&&haveAnArticle()||getArticlesCount()==0){
            throw new IndexOutOfSomeBoundsException();
        }else{
            throw new NoSuchElementException();
        }
    }else{
        int failuresCount=0;
        int j;
        Index[] subindexes=getSubindexes();
        for(j=0;j<this.subindexesCount;j++){
            try{
                return subindexes[j].getArticle(arIndex,articleKeyLength,head);
            }catch(NoSuchElementException e){
            }catch(IndexOutOfSomeBoundsException e){
                failuresCount++;
            }
        }
        if(failuresCount<this.subindexesCount){
            throw new NoSuchElementException();
        }else{
            throw new IndexOutOfSomeBoundsException();
        }
    }
}

private int getArticleTextOffset(){
    if(articleOffset==NULL_INT_VALUE){
        dataInput.seek(offset+PieceOfDictionary.INT_LENGTH);
        try{
            articleOffset=dataInput.readInt();
        }catch(IOException e){
            PieceOfDictionary.showErrorAndExit(Language.
                    getCurrentLanguageInscription("no_article_offset_error"));
        }
    }
    return articleOffset;
}

private int getArticlesCount(){
    if(articlesCount==NULL_INT_VALUE){
        dataInput.seek(offset);
        try{
            articlesCount=dataInput.readInt();
        }catch(IOException e){
            PieceOfDictionary.showErrorAndExit(Language.
                    getCurrentLanguageInscription("no_articles_count_error"));
        }
    }
    return articlesCount;
}

public String getKey(){
    return key==null?key=appendKey(new StringBuffer()).toString():key;
}

private Index getFreeSubindex(){
    getSubindexes();
    return freeSubindex;
}

public Head getHead(){
    return new Head(this,getKey());
}

private int getNextLevelArticlesCount(){
    getSubindexes();
    return nextLevelArticlesCount;
}

/**
 * Заполняет {@link Index#subindexes}, впоследствии сортируя его.
 */
private Index[] getSubindexes(){
    if(this.subindexes==null){
        Index[] subindexes=new Index[this.subindexesCount];
        int subindexOffset=getSubindexesOffset();
        if(subindexOffset==NULL_POINTER){
            return subindexes;
        }
        int j;
        Index subindex;
        int subindexCharacterIndex;
        boolean freshFreeSubindexNeeds=(subindexOffset&
                Index.HAVE_NOT_FREE_SUBINDEX_MASK)==0;
        subindexOffset&=Index.SUBINDEXES_OFFSET_MASK;
        for(j=0;j<this.subindexesCount;j++){
            dataInput.seek(this.offset+PieceOfDictionary.INT_LENGTH*3+
                    PieceOfDictionary.BYTE_LENGTH*2*j);
            try{
                subindexCharacterIndex=dataInput.readUnsignedByte();
                subindex=subindexes[j]=new Index(subindexOffset,subindexCharacterIndex==0xFF
                        ?FREE_CHARACTER:alphabet[subindexCharacterIndex],
                        dataInput.readUnsignedByte(),this);
                subindexOffset+=subindex.length();
                if(subindex.haveAnArticle()){
                    nextLevelArticlesCount++;
                }
                freshFreeSubindexNeeds=freshFreeSubindexNeeds&&
                        subindexCharacterIndex!=0xFF;
            }catch(IOException e){
                PieceOfDictionary.showErrorAndExit(Language.
                        getCurrentLanguageInscription("no_subindex_error"));
            }
        }
        if(freshFreeSubindexNeeds){
            PieceOfDictionary.showError(Language.getCurrentLanguageInscription(
                    "no_free_subindex_error"));
        }
        PieceOfDictionary.sort(subindexes);
        this.subindexes=subindexes;
    }
    return subindexes;
}

private int getSubindexesOffset(){
    if(subindexesOffset==NULL_INT_VALUE){
        dataInput.seek(offset+PieceOfDictionary.INT_LENGTH*2);
        try{
            subindexesOffset=dataInput.readInt();
        }catch(IOException e){
            PieceOfDictionary.showErrorAndExit(Language.
                    getCurrentLanguageInscription("no_subindexes_offset_error"));
        }
    }
    return subindexesOffset;
}

private boolean haveAnArticle(){
    return getArticleTextOffset()!=NULL_POINTER;
}

private int length(){
    return PieceOfDictionary.INT_LENGTH*3+PieceOfDictionary.BYTE_LENGTH*2*
            subindexesCount;
}

}

}
