
/**
 * Вектор, количество элементов которого ограничено заданным значением.
 *
 * @author quasiyoke
 */
public class LimitedVector{

private final int SIZE_LIMIT;

private Object[] objects;

private int objectsCount;

private boolean sizeLimitReached;

private int vectorStart;

/**
 * @throws IllegalArgumentException Если <code>sizeLimit</code> -- не
 * натуральное число.
 */
public LimitedVector(int sizeLimit){
    if(sizeLimit<=0){
        throw new IllegalArgumentException();
    }
    SIZE_LIMIT=sizeLimit;
    objects=new Object[SIZE_LIMIT];
    vectorStart=0;
}

/**
 * @param index Номер элемента, который требуется получить. Последний
 * добавленный элемент получает индекс 0.
 * @return Элемент с заданным номером или <code>null</code>, если
 * <code>index&lt;0||index&gt;=size()</code>.
 */
public Object elementAt(int index){
    return index<objectsCount&&index>=0?objects[(index=objectsCount-index-1)+
            vectorStart<SIZE_LIMIT?index+vectorStart:index+vectorStart-
            SIZE_LIMIT]:null;
}

/**
 * Добавляет ещё один элемент, который получит индекс 0. Если вектор уже достиг
 * своего максимального размера, теряется наиболее давно добавленный в него
 * элемент.
 */
public void addElement(Object object){
    if(sizeLimitReached){
        objects[vectorStart++]=object;
        if(vectorStart==SIZE_LIMIT){
            vectorStart=0;
        }
    }else{
        objects[objectsCount++]=object;
        sizeLimitReached=objectsCount>=SIZE_LIMIT;
    }
}

public int size(){
    return objectsCount;
}

}
